import matplotlib.colors as colors
import matplotlib.pyplot as plt
import numpy as np

data = {}

with open("results-nancy-gros.txt", "r") as f:
    for line in f:
        line = line.strip()

        left, right = line.split(":")
        nb_machines, nb_processes, nb_refs_per_process = [int(x) for x in left.split()]
        time = float(right.strip())

        if nb_machines not in data:
            data[nb_machines] = []
        data[nb_machines].append((nb_processes, nb_refs_per_process, time))

# Compute the relative difference between the two experiments
# max_diff = 0

# for (p1, r1, t1), (p2, r2, t2) in zip(data[3], data[5]):
#     assert p1 == p2 and r1 == r2
#     max_diff = max(max_diff, abs(t1 - t2) / t1)

# print(f"Max diff: {max_diff:.2%}")

for a, values in data.items():
    ps = sorted(set(p for p, r, t in values))
    rs = sorted(set(r for p, r, t in values))

    grid = np.full((len(rs), len(ps)), np.nan)

    for p, r, t in values:
        i = rs.index(r)
        j = ps.index(p)
        grid[i, j] = t / (p * r)

    fig, ax = plt.subplots()

    ax.imshow(grid, aspect="auto", origin="lower", norm=colors.LogNorm())

    cbar = ax.figure.colorbar(ax.images[0], ax=ax)
    cbar.set_label("Time per iteration per reference (ms)")
    plt.xticks(np.arange(len(ps)), [str(int(p)) for p in ps])
    plt.yticks(np.arange(len(rs)), [str(int(r)) for r in rs])
    plt.xlabel("Number of processes")
    plt.ylabel("Number of references sent by process")
    plt.tight_layout()

    plt.savefig(f"heatmap_{a}.png")

    for (i, j), val in np.ndenumerate(grid):
        ax.text(j, i, f"{val:.2f}", ha="center", va="center", color="white")

    plt.savefig(f"heatmap_{a}_annotated.png")



import time

import numpy as np
import pytest
import ray


@pytest.fixture
def ray_cluster():
    """Start a Ray cluster for this test"""
    ray.init()
    yield
    ray.shutdown()


def wait_for_head_node() -> None:
    """Wait until the head node is ready"""
    while True:
        try:
            ray.get_actor("simulation_head", namespace="doreisa")
            return
        except ValueError:
            time.sleep(0.1)


@ray.remote(num_cpus=0, max_retries=0)
def simple_worker(
    *,
    rank: int,
    position: tuple[int, ...],
    chunks_per_dim: tuple[int, ...],
    nb_chunks_of_node: int,
    chunk_size: tuple[int, ...],
    nb_iterations: int,
    node_id: str | None = None,
    array_name: str = "array",
    dtype: np.dtype = np.int32,  # type: ignore
) -> None:
    """Worker node sending chunks of data"""
    from doreisa.simulation_node import Client

    client = Client(_fake_node_id=node_id)

    array = (rank + 1) * np.ones(chunk_size, dtype=dtype)

    for i in range(nb_iterations):
        client.add_chunk(array_name, position, chunks_per_dim, nb_chunks_of_node, i, i * array, store_externally=False)



import random
import time
from collections import Counter

import ray
from dask.core import get_dependencies

from doreisa._scheduling_actor import ChunkRef, ScheduledByOtherActor


def doreisa_get(dsk, keys, **kwargs):
    debug_logs_path: str | None = kwargs.get("doreisa_debug_logs", None)

    def log(message: str, debug_logs_path: str | None) -> None:
        if debug_logs_path is not None:
            with open(debug_logs_path, "a") as f:
                f.write(f"{time.time()} {message}\n")

    log("1. Begin Doreisa scheduler", debug_logs_path)

    # Sort the graph by keys to make scheduling deterministic
    dsk = {k: v for k, v in sorted(dsk.items())}

    head_node = ray.get_actor("simulation_head", namespace="doreisa")  # noqa: F841

    # TODO this will not work all the time
    assert isinstance(keys, list) and len(keys) == 1
    if isinstance(keys[0], list):
        assert len(keys[0]) == 1
        key = keys[0][0]
    else:
        key = keys[0]

    # Find the scheduling actors
    scheduling_actors = ray.get(head_node.list_scheduling_actors.remote())

    # Find a not too bad scheduling strategy
    # Good scheduling in a tree
    partition = {k: -1 for k in dsk.keys()}

    # def explore(key, v: int):
    #     # Only works for trees for now
    #     assert scheduling[key] == -1
    #     scheduling[key] = v
    #     for dep in get_dependencies(dsk, key):
    #         explore(dep, v)

    # scheduling[key] = 0
    # c = 0
    # for dep1 in get_dependencies(dsk, key):
    #     scheduling[dep1] = 0

    #     for dep2 in get_dependencies(dsk, dep1):
    #         scheduling[dep2] = 0

    #         for dep3 in get_dependencies(dsk, dep2):
    #             scheduling[dep3] = 0

    #             for dep4 in get_dependencies(dsk, dep3):
    #                 scheduling[dep4] = 0

    #                 for dep5 in get_dependencies(dsk, dep4):
    #                     explore(dep5, c % len(scheduling_actors))
    #                     c += 1

    # assert -1 not in scheduling.values()

    # scheduling = {k: randint(0, len(scheduling_actors) - 1) for k in dsk.keys()}
    # scheduling = {k: i % len(scheduling_actors) for i, k in enumerate(dsk.keys())}

    # Make sure the leafs are scheduled on the right actor
    # for key, val in dsk.items():
    #     match val:
    #         case ("doreisa_chunk", actor_id):
    #             scheduling[key] = actor_id
    #         case _:
    #             pass

    def explore(k) -> int:
        val = dsk[k]

        if isinstance(val, ChunkRef):
            partition[k] = val.actor_id
        else:
            actors_dependencies = [explore(dep) for dep in get_dependencies(dsk, k)]

            if not actors_dependencies:
                # The task is a leaf, we use a random actor
                partition[k] = random.randint(0, len(scheduling_actors) - 1)
            else:
                partition[k] = Counter(actors_dependencies).most_common(1)[0][0]

        return partition[k]

    explore(key)

    log("2. Graph partitionning done", debug_logs_path)

    partitionned_graphs: dict[int, dict] = {actor_id: {} for actor_id in range(len(scheduling_actors))}

    for k, v in dsk.items():
        actor_id = partition[k]

        partitionned_graphs[actor_id][k] = v

        for dep in get_dependencies(dsk, k):
            if partition[dep] != actor_id:
                partitionned_graphs[actor_id][dep] = ScheduledByOtherActor(partition[dep])

    log("3. Partitionned graphs created", debug_logs_path)

    graph_id = random.randint(0, 2**128 - 1)

    for id, actor in enumerate(scheduling_actors):
        if partitionned_graphs[id]:
            actor.schedule_graph.remote(graph_id, partitionned_graphs[id])

    log("4. Graph scheduled", debug_logs_path)

    res_ref = scheduling_actors[partition[key]].get_value.remote(graph_id, key)

    if kwargs.get("ray_persist"):
        if isinstance(keys[0], list):
            return [[res_ref]]
        return [res_ref]

    res = ray.get(ray.get(res_ref))

    log("5. End Doreisa scheduler", debug_logs_path)

    if isinstance(keys[0], list):
        return [[res]]
    return [res]



import dask.array as da
import ray

from tests.utils import ray_cluster, simple_worker, wait_for_head_node  # noqa: F401

NB_ITERATIONS = 10


@ray.remote(max_retries=0)
def head() -> None:
    """The head node checks that the values are correct"""
    from doreisa.head_node import init
    from doreisa.window_api import ArrayDefinition, run_simulation

    init()

    def simulation_callback(array: list[da.Array], timestep: int):
        if timestep == 0:
            assert len(array) == 1
            return

        assert array[0].sum().compute() == 10 * (timestep - 1)
        assert array[1].sum().compute() == 10 * timestep

        # Test a computation where the two arrays are used at the same time.
        # This checks that they are defined with different names.
        assert (array[1] - array[0]).sum().compute() == 10

    run_simulation(
        simulation_callback,
        [
            ArrayDefinition("array", window_size=2),
        ],
        max_iterations=NB_ITERATIONS,
    )


def test_sliding_window(ray_cluster) -> None:  # noqa: F811
    head_ref = head.remote()
    wait_for_head_node()

    worker_refs = []
    for rank in range(4):
        worker_refs.append(
            simple_worker.remote(
                rank=rank,
                position=(rank // 2, rank % 2),
                chunks_per_dim=(2, 2),
                nb_chunks_of_node=1,
                chunk_size=(1, 1),
                nb_iterations=NB_ITERATIONS,
                node_id=f"node_{rank}",
            )
        )

    ray.get([head_ref] + worker_refs)



import asyncio
from typing import Generic, TypeVar

K = TypeVar("K")
V = TypeVar("V")


class AsyncDict(Generic[K, V]):
    def __init__(self) -> None:
        self._data: dict[K, V] = {}
        self._new_key_event = asyncio.Event()

    def __setitem__(self, key: K, value: V):
        self._data[key] = value
        self._new_key_event.set()
        self._new_key_event.clear()

    def __getitem__(self, key: K) -> V:
        return self._data[key]

    async def wait_for_key(self, key: K) -> V:
        while key not in self._data:
            await self._new_key_event.wait()
        return self._data[key]

    def __contains__(self, key: K) -> bool:
        return key in self._data

    def __len__(self) -> int:
        return len(self._data)



import asyncio
import pickle
from dataclasses import dataclass

import numpy as np
import ray
import ray.actor
import ray.util.dask.scheduler

from doreisa import Timestep
from doreisa._async_dict import AsyncDict


@dataclass
class ChunkRef:
    """
    Represents a chunk of an array in a Dask task graph.

    The task corresping to this object must be scheduled by the actor who has the actual
    data. This class is used since Dask tends to inline simple tuples. This may change
    in newer versions of Dask.
    """

    actor_id: int
    array_name: str  # The real name, without the timestep
    timestep: Timestep
    position: tuple[int, ...]

    # Set for one chunk only.
    _all_chunks: ray.ObjectRef | None = None


@dataclass
class ScheduledByOtherActor:
    """
    Represents a task that is scheduled by another actor in the part of the task graph sent to an actor.
    """

    actor_id: int


class GraphInfo:
    """
    Information about graphs and their scheduling.
    """

    def __init__(self):
        self.scheduled_event = asyncio.Event()
        self.refs: dict[str, ray.ObjectRef] = {}


@ray.remote(num_cpus=0, enable_task_events=False)
def patched_dask_task_wrapper(func, repack, key, ray_pretask_cbs, ray_posttask_cbs, *args, first_call=True):
    """
    Patched version of the original dask_task_wrapper function.

    This version received ObjectRefs first, and calls itself a second time to unwrap the ObjectRefs.
    The result is an ObjectRef.

    TODO can probably be rewritten without copying the whole function
    """

    if first_call:
        assert all([isinstance(a, ray.ObjectRef) for a in args])
        # Use one CPU for the actual computation
        return patched_dask_task_wrapper.options(num_cpus=1).remote(
            func, repack, key, ray_pretask_cbs, ray_posttask_cbs, *args, first_call=False
        )

    if ray_pretask_cbs is not None:
        pre_states = [cb(key, args) if cb is not None else None for cb in ray_pretask_cbs]
    repacked_args, repacked_deps = repack(args)
    # Recursively execute Dask-inlined tasks.
    actual_args = [ray.util.dask.scheduler._execute_task(a, repacked_deps) for a in repacked_args]
    # Execute the actual underlying Dask task.
    result = func(*actual_args)

    if ray_posttask_cbs is not None:
        for cb, pre_state in zip(ray_posttask_cbs, pre_states):
            if cb is not None:
                cb(key, result, pre_state)

    return result


@ray.remote(num_cpus=0, enable_task_events=False)
def remote_ray_dask_get(dsk, keys):
    import ray.util.dask

    # Monkey-patch Dask-on-Ray
    ray.util.dask.scheduler.dask_task_wrapper = patched_dask_task_wrapper

    return ray.util.dask.ray_dask_get(dsk, keys, ray_persist=True)


class _ArrayTimestep:
    def __init__(self):
        # Triggered when all the chunks are ready
        self.chunks_ready_event: asyncio.Event = asyncio.Event()

        # {position: chunk}
        self.local_chunks: AsyncDict[tuple[int, ...], ray.ObjectRef | bytes] = AsyncDict()


class _Array:
    def __init__(self):
        # Indicates if set_owned_chunks method has been called for this array.
        self.is_registered = False

        # Chunks owned by this actor for this array.
        # {(chunk position, chunk size), ...}
        self.owned_chunks: set[tuple[tuple[int, ...], tuple[int, ...]]] = set()

        self.timesteps: AsyncDict[Timestep, _ArrayTimestep] = AsyncDict()


@ray.remote
class SchedulingActor:
    """
    Actor in charge of gathering ObjectRefs and scheduling the tasks produced by the head node.
    """

    def __init__(self, actor_id: int) -> None:
        self.actor_id = actor_id
        self.actor_handle = ray.get_runtime_context().current_actor

        self.head = ray.get_actor("simulation_head", namespace="doreisa")
        self.scheduling_actors: list[ray.actor.ActorHandle] = []

        # For collecting chunks
        self.arrays: AsyncDict[str, _Array] = AsyncDict()

        # For scheduling
        self.graph_infos: AsyncDict[int, GraphInfo] = AsyncDict()

    def ready(self) -> None:
        pass

    def _pack_object_ref(self, refs: list[ray.ObjectRef]):
        """
        Used to create an ObjectRef containing the given ObjectRef.
        This allows having the expected format in the task graph.

        This is a method instead of a function with `num_cpus=0` to avoid starting many
        new workers.
        """
        return refs[0]

    async def add_chunk(
        self,
        array_name: str,
        timestep: int,
        chunk_position: tuple[int, ...],
        dtype: np.dtype,
        nb_chunks_per_dim: tuple[int, ...],
        nb_chunks_of_node: int,
        chunk: list[ray.ObjectRef],
        chunk_shape: tuple[int, ...],
    ) -> None:
        if array_name not in self.arrays:
            self.arrays[array_name] = _Array()
        array = self.arrays[array_name]

        if timestep not in array.timesteps:
            array.timesteps[timestep] = _ArrayTimestep()
        array_timestep = array.timesteps[timestep]

        assert chunk_position not in array_timestep.local_chunks
        array_timestep.local_chunks[chunk_position] = self.actor_handle._pack_object_ref.remote(chunk)

        array.owned_chunks.add((chunk_position, chunk_shape))

        if len(array_timestep.local_chunks) == nb_chunks_of_node:
            if not array.is_registered:
                # Register the array with the head node
                await self.head.set_owned_chunks.options(enable_task_events=False).remote(
                    self.actor_id,
                    array_name,
                    dtype,
                    nb_chunks_per_dim,
                    list(array.owned_chunks),
                )
                array.is_registered = True

            chunks = []
            for position, size in array.owned_chunks:
                c = array_timestep.local_chunks[position]
                assert isinstance(c, ray.ObjectRef)
                chunks.append(c)
                array_timestep.local_chunks[position] = pickle.dumps(c)

            all_chunks_ref = ray.put(chunks)

            await self.head.chunks_ready.options(enable_task_events=False).remote(
                array_name, timestep, [all_chunks_ref]
            )

            array_timestep.chunks_ready_event.set()
            array_timestep.chunks_ready_event.clear()
        else:
            await array_timestep.chunks_ready_event.wait()

    async def schedule_graph(self, graph_id: int, dsk: dict) -> None:
        # Find the scheduling actors
        if not self.scheduling_actors:
            self.scheduling_actors = await self.head.list_scheduling_actors.options(enable_task_events=False).remote()

        info = GraphInfo()
        self.graph_infos[graph_id] = info

        for key, val in dsk.items():
            # Adapt external keys
            if isinstance(val, ScheduledByOtherActor):
                actor = self.scheduling_actors[val.actor_id]
                dsk[key] = actor.get_value.options(enable_task_events=False).remote(graph_id, key)

            # Replace the false chunks by the real ObjectRefs
            if isinstance(val, ChunkRef):
                assert val.actor_id == self.actor_id

                array = await self.arrays.wait_for_key(val.array_name)
                array_timestep = await array.timesteps.wait_for_key(val.timestep)
                ref = await array_timestep.local_chunks.wait_for_key(val.position)

                if isinstance(ref, bytes):  # This may not be the case depending on the asyncio scheduling order
                    ref = pickle.loads(ref)
                else:
                    ref = pickle.loads(pickle.dumps(ref))  # To free the memory automatically

                dsk[key] = ref

        # We will need the ObjectRefs of these keys
        keys_needed = list(dsk.keys())

        refs = await remote_ray_dask_get.remote(dsk, keys_needed)

        for key, ref in zip(keys_needed, refs):
            info.refs[key] = ref

        info.scheduled_event.set()

    async def get_value(self, graph_id: int, key: str):
        graph_info = await self.graph_infos.wait_for_key(graph_id)

        await graph_info.scheduled_event.wait()
        return await graph_info.refs[key]



import dask.array as da
import numpy as np
import ray

from tests.utils import ray_cluster, simple_worker, wait_for_head_node  # noqa: F401


@ray.remote(max_retries=0)
def head_script() -> None:
    """The head node checks that the values are correct"""
    from doreisa.head_node import init
    from doreisa.window_api import ArrayDefinition, run_simulation

    init()

    def simulation_callback(array: da.Array, timestep: int):
        assert array.dtype == np.int8

    run_simulation(
        simulation_callback,
        [ArrayDefinition("array")],
        max_iterations=1,
    )


def test_dtype(ray_cluster) -> None:  # noqa: F811
    head_ref = head_script.remote()
    wait_for_head_node()

    worker_ref = simple_worker.remote(
        rank=0,
        position=(0,),
        chunks_per_dim=(1,),
        nb_chunks_of_node=1,
        chunk_size=(1,),
        nb_iterations=1,
        node_id="node",
        dtype=np.int8,
    )

    ray.get([head_ref, worker_ref])



import logging
from typing import Callable

import numpy as np
import ray
import ray.actor


class Client:
    """
    Used by the MPI nodes to send data to the analytic cluster.

    The client is in charge of a several chunks of data. Each chunk is at a certain position in an
    array.
    """

    def __init__(self, *, _fake_node_id: str | None = None) -> None:
        """
        Args:
            _fake_node_id: The ID of the node. If None, the ID is taken from the Ray runtime context.
                This is useful for testing with several scheduling actors on a single machine.
        """
        if not ray.is_initialized():
            ray.init(address="auto", log_to_driver=False, logging_level=logging.ERROR)

        self.node_id = _fake_node_id or ray.get_runtime_context().get_node_id()

        self.head = ray.get_actor("simulation_head", namespace="doreisa")
        self.scheduling_actor: ray.actor.ActorHandle = ray.get(
            self.head.scheduling_actor.remote(self.node_id, is_fake_id=bool(_fake_node_id))
        )

        self.preprocessing_callbacks: dict[str, Callable] = ray.get(self.head.preprocessing_callbacks.remote())

    def add_chunk(
        self,
        array_name: str,
        chunk_position: tuple[int, ...],
        nb_chunks_per_dim: tuple[int, ...],
        nb_chunks_of_node: int,
        timestep: int,
        chunk: np.ndarray,
        store_externally: bool = False,
    ) -> None:
        """
        Make a chunk of data available to the analytic.

        Args:
            array_name: The name of the array.
            chunk_position: The position of the chunk in the array.
            nb_chunks_per_dim: The number of chunks per dimension.
            nb_chunks_of_node: The number of chunks sent by this node. The scheduling actor will
                inform the head actor when all the chunks are ready.
            chunk: The chunk of data.
            store_externally: If True, the data is stored externally. TODO Not implemented yet.
        """
        chunk = self.preprocessing_callbacks[array_name](chunk)

        # Setting the owner allows keeping the reference when the simulation script terminates.
        ref = ray.put(chunk, _owner=self.scheduling_actor)

        future: ray.ObjectRef = self.scheduling_actor.add_chunk.options(enable_task_events=False).remote(
            array_name,
            timestep,
            chunk_position,
            chunk.dtype,
            nb_chunks_per_dim,
            nb_chunks_of_node,
            [ref],
            chunk.shape,
        )  # type: ignore

        # Wait until the data is processed before returning to the simulation
        ray.get(future)



import dask.array as da
import ray

from tests.utils import ray_cluster, simple_worker, wait_for_head_node  # noqa: F401

NB_ITERATIONS = 10


@ray.remote(max_retries=0)
def head_script() -> None:
    """The head node checks that the values are correct"""
    from doreisa.head_node import init
    from doreisa.window_api import ArrayDefinition, run_simulation

    init()

    def simulation_callback(array: da.Array, timestep: int):
        # This is the standard dask task graph
        assert len(array.sum().dask) == 9

        x = array.sum().persist()

        # We still have a dask array
        assert isinstance(x, da.Array)

        # But only one task in the task graph, since the result is being computed
        assert len(x.dask) == 1

        x_final = x.compute()
        assert x_final == 10 * timestep

    run_simulation(
        simulation_callback,
        [ArrayDefinition("array")],
        max_iterations=NB_ITERATIONS,
    )


def test_dask_persist(ray_cluster) -> None:  # noqa: F811
    head_ref = head_script.remote()
    wait_for_head_node()

    worker_refs = []
    for rank in range(4):
        worker_refs.append(
            simple_worker.remote(
                rank=rank,
                position=(rank // 2, rank % 2),
                chunks_per_dim=(2, 2),
                nb_chunks_of_node=1,
                chunk_size=(1, 1),
                nb_iterations=NB_ITERATIONS,
                node_id=f"node_{rank}",
            )
        )

    ray.get([head_ref] + worker_refs)



"""
Run an experiment to dertermine the maximum number of references that a single node can process.

This script must be executed from the root of the repository.
"""

import threading
import time

import execo
import execo_g5k


def run_experiment(nb_reserved_nodes: int, nb_workers: int, nb_chunks_sent: int) -> None:
    """
    Params:
        nb_reserved_nodes: The number of nodes to reserve on the Grid'5000 platform
        nb_workers: The number of MPI processes to use
        nb_chunks_sent: The number of chunks to send to the head node by each worker at each iteration
    """

    print("Starting experiment with", nb_reserved_nodes, nb_workers, nb_chunks_sent)

    # Reserve the resources
    jobs = execo_g5k.oarsub(
        [
            (
                execo_g5k.OarSubmission(resources=f"{{cluster='gros'}}/nodes={nb_reserved_nodes}", walltime=6 * 60),
                "nancy",
            )
        ]
    )
    job_id, site = jobs[0]

    # Get the nodes to use
    nodes = execo_g5k.get_oar_job_nodes(job_id, site)
    head_node, nodes = nodes[0], nodes[1:]

    print(head_node, nodes)

    # Start the head node
    head_node_cmd = execo.SshProcess(
        'ulimit -n 65535; singularity exec doreisa/docker/images/doreisa-simulation.sif bash -c "cd doreisa; ray start --head --port=4242; python3 experiments/refs-gathering-benchmark/head.py"',
        head_node,
    )
    head_node_cmd.start()

    time.sleep(5)

    print("Head node started")

    # Start the simulation nodes
    for node in nodes:
        node_cmd = execo.SshProcess(
            f"""ulimit -n 65535; singularity exec doreisa/docker/images/doreisa-simulation.sif bash -c "ray start --address='{head_node.address}:4242'; sleep infinity" """,
            node,
        )
        node_cmd.start()

    # Wait for everything to start
    time.sleep(10)

    workers = []

    while len(workers) < nb_workers:
        node = nodes[len(workers) % len(nodes)]

        worker = execo.SshProcess(
            f"ulimit -n 65535; cd doreisa && singularity exec docker/images/doreisa-simulation.sif python3 experiments/refs-gathering-benchmark/worker.py {len(workers)} {nb_workers} {nb_chunks_sent}",
            node,
        )
        worker.start()
        workers.append(worker)

    for worker in workers:
        worker.wait()

    # Release the ressources
    execo_g5k.oardel(jobs)


threads = []

# Demonstrate the bottleneck when sending the references one by one.
# The results should be the same with 2 and 4 simulation nodes: the head node is the bottleneck.
for i in range(10):
    for j in range(10):
        if i + j > 12:
            continue

        # A head and 4 simulation nodes
        threads.append(threading.Thread(target=run_experiment, args=(1 + 4, 2**i, 2**j)))
        threads[-1].start()

for t in threads:
    t.join()
